@@ -100,8 +100,9 @@ def test_transfer_not_contiguous():
 @guard_devsup
 def transfer_not_contiguous(shp, dtype):
     a = numpy.random.rand(*shp) * 10
-    a = a[::-1]
     b = pygpu.array(a, context=ctx)
+    a = a[::-1]
+    b = b[::-1]
     c = numpy.asarray(b)
 
     assert numpy.allclose(c, a)
@@ -121,11 +122,12 @@ def test_transfer_fortran():
 @guard_devsup
 def transfer_fortran(shp, dtype):
     a = numpy.random.rand(*shp) * 10
+    b = pygpu.array(a, context=ctx)
     a_ = numpy.asfortranarray(a)
     if len(shp) > 1:
         assert a_.strides != a.strides
     a = a_
-    b = pygpu.array(a, context=ctx)
+    b = pygpu.asfortranarray(b)
     c = numpy.asarray(b)
 
     assert a.shape == b.shape == c.shape
@@ -193,7 +195,7 @@ def asfortranarray(shp, dtype, offseted_outer, offseted_inner, sliced, order):
 
     # numpy upcast with a view to 1d scalar.
     if gpu.flags['F_CONTIGUOUS']:
-        assert b.gpudata == gpu.gpudata
+        assert ctx.kind != b'cuda' or b.gpudata == gpu.gpudata
     elif (sliced != 1 or shp == () or (offseted_outer and len(shp) > 1) or
           (order != 'f' and len(shp) > 1)):
         assert b is not gpu
@@ -284,7 +286,8 @@ def test_mapping_getitem_ellipsis():
 def mapping_getitem_ellipsis(shp, dtype, offseted):
     a, a_gpu = gen_gpuarray(shp, dtype, offseted, ctx=ctx)
     b = a_gpu[...]
-    assert b.gpudata == a_gpu.gpudata
+    if ctx.kind == b'cuda':
+        assert b.gpudata == a_gpu.gpudata
     assert b.strides == a.strides
     assert b.shape == a.shape
     b_cpu = numpy.asarray(b)
@@ -811,8 +814,9 @@ class TestPickle(unittest.TestCase):
             pickle.dumps(ctx, protocol=-1)
 
     def test_GpuKernel(self):
-        k = GpuKernel("KERNEL void nothing(GLOBAL_MEM ga_float *in) "
-                      "{in[0] = 0;}", "nothing", [], context=ctx)
+        k = GpuKernel("#include \"cluda.h\"\nKERNEL void "
+                      "k(GLOBAL_MEM ga_float *in)"
+                      "{in[0] = 0;}", "k", [], context=ctx)
         with self.assertRaises(RuntimeError):
             pickle.dumps(k)
         with self.assertRaises(RuntimeError):
