@@ -268,11 +268,6 @@ GPUARRAY_PUBLIC int GpuArray_fromdata(GpuArray *a,
                                       const size_t *dims,
                                       const ssize_t *strides, int writeable);
 
-GPUARRAY_PUBLIC int GpuArray_copy_from_host(GpuArray *a,
-                                            gpucontext *ctx, void *buf, int typecode,
-                                            unsigned int nd, const size_t *dims,
-                                            const ssize_t *strides);
-
 /**
  * Initialize an array structure to provide a view of another.
  *
