@@ -53,7 +53,8 @@ static const char *estr(cublasStatus_t err) {
 
 static inline int error_cublas(error *e, const char *msg, cublasStatus_t err) {
   return error_fmt(e, (err == CUBLAS_STATUS_ARCH_MISMATCH) ? GA_DEVSUP_ERROR : GA_BLAS_ERROR,
-                   "%s: %s", msg, estr(err));
+                   "%s: %s%s", msg, estr(err),
+		   err == CUBLAS_STATUS_NOT_INITIALIZED ? " (Possibly because the driver version is too old for the cuda version)" : "");
 }
 
 #define CUBLAS_EXIT_ON_ERROR(ctx, cmd) do {       \
@@ -72,15 +73,17 @@ typedef struct _blas_handle {
   GpuKernel dgemvBH_T_a1_b1_small;
   GpuKernel sgerBH_gen_small;
   GpuKernel dgerBH_gen_small;
+  uint8_t tensorCore;
 } blas_handle;
 
 #define LARGE_VAL(v) (v >= INT_MAX)
 
 static const char *code_sgemvBH_N_a1_b1_small =                         \
-  "extern \"C\"__global__ void sgemv(const float *A[], size_t lda, "    \
-  "                                  const float *x[], size_t incx, "   \
-  "                                  float *y[], size_t incy, "         \
-  "                                  size_t b, size_t m, size_t n) {"   \
+  "#include \"cluda.h\"\n"                                              \
+  "KERNEL void sgemv(const float *A[], size_t lda, "                    \
+  "                  const float *x[], size_t incx, "                   \
+  "                  float *y[], size_t incy, "                         \
+  "                  size_t b, size_t m, size_t n) {"                   \
   "  for (size_t p = blockIdx.y * blockDim.y + threadIdx.y; p < b;"     \
   "       p += gridDim.y * blockDim.y) {"                               \
   "    for (size_t i = blockIdx.x * blockDim.x + threadIdx.x; i < m;"   \
@@ -94,51 +97,37 @@ static const char *code_sgemvBH_N_a1_b1_small =                         \
   "        Ap += lda;"                                                  \
   "        xp += incx;"                                                 \
   "      }"                                                             \
-  "     atomicAdd(&y[p][i*incy], yi);"                                  \
+  "     atom_add_fg(&y[p][i*incy], yi);"                                \
   "    }"                                                               \
   "  }"                                                                 \
   "}\n";
 
-static const char *code_sgemvBH_T_a1_b1_small =                         \
-  "extern \"C\" __global__ void sgemv(const float *A[], size_t lda, "   \
-  "                                   const float *x[], size_t incx, "  \
-  "                                   float *y[], size_t incy, "        \
-  "                                   size_t b, size_t m, size_t n) {"  \
-  "  size_t i = blockIdx.x * blockDim.x + threadIdx.x;"                 \
-  "  size_t p = blockIdx.y * blockDim.y + threadIdx.y;"                 \
-  "  if (i >= m || p >= b) return;"                                     \
-  "  float yi = 0.0f;"                                                  \
-  "  const float *Ap = A[p] + i * lda;"                                 \
-  "  const float *xp = x[p];\n"                                         \
-  "  # pragma unroll 32\n"                                              \
-  "  for (size_t j = 0; j < n; j++) {"                                  \
-  "    yi += Ap[j] * xp[0];"                                            \
-  "    xp += incx;"                                                     \
-  "  }"                                                                 \
-  "  atomicAdd(&y[p][i*incy], yi);"                                     \
+static const char *code_sgemvBH_T_a1_b1_small =         \
+  "#include \"cluda.h\"\n"                              \
+  "KERNEL void sgemv(const float *A[], size_t lda, "    \
+  "                  const float *x[], size_t incx, "   \
+  "                  float *y[], size_t incy, "         \
+  "                  size_t b, size_t m, size_t n) {"   \
+  "  size_t i = blockIdx.x * blockDim.x + threadIdx.x;" \
+  "  size_t p = blockIdx.y * blockDim.y + threadIdx.y;" \
+  "  if (i >= m || p >= b) return;"                     \
+  "  float yi = 0.0f;"                                  \
+  "  const float *Ap = A[p] + i * lda;"                 \
+  "  const float *xp = x[p];\n"                         \
+  "  # pragma unroll 32\n"                              \
+  "  for (size_t j = 0; j < n; j++) {"                  \
+  "    yi += Ap[j] * xp[0];"                            \
+  "    xp += incx;"                                     \
+  "  }"                                                 \
+  "  atom_add_fg(&y[p][i*incy], yi);"                   \
   "}\n";
 
-static const char *atomicadd_double =                                   \
-  "#if __CUDA_ARCH__ < 600\n"						\
-  "__device__ double atomicAdd(double* address, double val) {"          \
-  "  unsigned long long int* address_as_ull ="                          \
-  "  (unsigned long long int*)address;"                                 \
-  "  unsigned long long int old = *address_as_ull, assumed;"            \
-  "  do {"                                                              \
-  "    assumed = old;"                                                  \
-  "    old = atomicCAS(address_as_ull, assumed,"                        \
-  "                    __double_as_longlong(val +"                      \
-  "                    __longlong_as_double(assumed)));"                \
-  "  } while (assumed != old);"                                         \
-  "  return __longlong_as_double(old);"                                 \
-  "}\n"									\
-  "#endif\n";
-
 static const char *code_dgemvBH_N_a1_b1_small =                         \
-  "extern \"C\" __global__ void dgemv(const double *A[], size_t lda, "  \
-  "                                   const double *x[], size_t incx, " \
-  "                                   double *y[], size_t incy, "       \
-  "                                   size_t b, size_t m, size_t n) {"  \
+  "#include \"cluda.h\"\n"                                              \
+  "KERNEL void dgemv(const double *A[], size_t lda, "                   \
+  "                  const double *x[], size_t incx, "                  \
+  "                  double *y[], size_t incy, "                        \
+  "                  size_t b, size_t m, size_t n) {"                   \
   "  for (size_t p = blockIdx.y * blockDim.y + threadIdx.y; p < b;"     \
   "       p += gridDim.y * blockDim.y) {"                               \
   "    for (size_t i = blockIdx.x * blockDim.x + threadIdx.x; i < m;"   \
@@ -152,32 +141,34 @@ static const char *code_dgemvBH_N_a1_b1_small =                         \
   "        Ap += lda;"                                                  \
   "        xp += incx;"                                                 \
   "      }"                                                             \
-  "     atomicAdd(&y[p][i*incy], yi);"                                  \
+  "     atom_add_dg(&y[p][i*incy], yi);"                                \
   "    }"                                                               \
   "  }"                                                                 \
   "}\n";
 
-static const char *code_dgemvBH_T_a1_b1_small =                         \
-  "extern \"C\" __global__ void dgemv(const double *A[], size_t lda, "  \
-  "                                   const double *x[], size_t incx, " \
-  "                                   double *y[], size_t incy, "       \
-  "                                   size_t b, size_t m, size_t n) {"  \
-  "  size_t i = blockIdx.x * blockDim.x + threadIdx.x;"                 \
-  "  size_t p = blockIdx.y * blockDim.y + threadIdx.y;"                 \
-  "  if (i >= m || p >= b) return;"                                     \
-  "  double yi = 0.0;"                                                  \
-  "  const double *Ap = A[p] + i * lda;"                                \
-  "  const double *xp = x[p];\n"                                        \
-  "  # pragma unroll 32\n"                                              \
-  "  for (size_t j = 0; j < n; j++) {"                                  \
-  "    yi += Ap[j] * xp[0];"                                            \
-  "    xp += incx;"                                                     \
-  "  }"                                                                 \
-  "  atomicAdd(&y[p][i*incy], yi);"                                     \
+static const char *code_dgemvBH_T_a1_b1_small =         \
+  "#include \"cluda.h\"\n"                              \
+  "KERNEL void dgemv(const double *A[], size_t lda, "   \
+  "                  const double *x[], size_t incx, "  \
+  "                  double *y[], size_t incy, "        \
+  "                  size_t b, size_t m, size_t n) {"   \
+  "  size_t i = blockIdx.x * blockDim.x + threadIdx.x;" \
+  "  size_t p = blockIdx.y * blockDim.y + threadIdx.y;" \
+  "  if (i >= m || p >= b) return;"                     \
+  "  double yi = 0.0;"                                  \
+  "  const double *Ap = A[p] + i * lda;"                \
+  "  const double *xp = x[p];\n"                        \
+  "  # pragma unroll 32\n"                              \
+  "  for (size_t j = 0; j < n; j++) {"                  \
+  "    yi += Ap[j] * xp[0];"                            \
+  "    xp += incx;"                                     \
+  "  }"                                                 \
+  "  atom_add_dg(&y[p][i*incy], yi);"                   \
   "}\n";
 
 static const char *code_sgerBH_gen_small =                              \
-  "extern \"C\" __global__ void _sgerBH_gen_small("                     \
+  "#include \"cluda.h\"\n"                                              \
+  "KERNEL void _sgerBH_gen_small("                                      \
   "    const float *x[], size_t incx,"                                  \
   "    const float *y[], size_t incy,"                                  \
   "    float alpha, float *A[], size_t lda,"                            \
@@ -186,13 +177,14 @@ static const char *code_sgerBH_gen_small =                              \
   "  size_t j = blockIdx.y * blockDim.y + threadIdx.y;"                 \
   "  if (i >= m || j >= n) return;"                                     \
   "  for (size_t p = blockIdx.z; p < b; p += gridDim.z) {"              \
-  "    atomicAdd(&A[p][j * lda + i],"                                   \
-  "              alpha * x[p][i * incx] * y[p][j * incy]);"             \
+  "    atom_add_fg(&A[p][j * lda + i],"                                 \
+  "                alpha * x[p][i * incx] * y[p][j * incy]);"           \
   "  }"                                                                 \
   "}\n";
 
 static const char *code_dgerBH_gen_small =                              \
-  "extern \"C\" __global__ void _dgerBH_gen_small("                     \
+  "#include \"cluda.h\"\n"                                              \
+  "KERNEL void _dgerBH_gen_small("                                      \
   "      const double *x[], size_t incx, "                              \
   "      const double *y[], size_t incy,"                               \
   "      double alpha, double *A[], size_t lda,"                        \
@@ -201,17 +193,18 @@ static const char *code_dgerBH_gen_small =                              \
   "  size_t j = blockIdx.y * blockDim.y + threadIdx.y;"                 \
   "  if (i >= m || j >= n) return;"                                     \
   "  for (size_t p = blockIdx.z; p < b; p += gridDim.z) {"              \
-  "    atomicAdd(&A[p][j * lda + i],"                                   \
-  "              alpha * x[p][i * incx] * y[p][j * incy]);"             \
+  "    atom_add_dg(&A[p][j * lda + i],"                                 \
+  "                alpha * x[p][i * incx] * y[p][j * incy]);"           \
   "  }"                                                                 \
   "}\n";
 
 static int setup(gpucontext *c) {
   cuda_context *ctx = (cuda_context *)c;
   blas_handle *handle;
-  const char *tmp[2];
+  CUdevice dev;
   cublasStatus_t err;
   int types[10];
+  int major, minor;
   int e;
 
   if (ctx->blas_handle != NULL)
@@ -222,6 +215,23 @@ static int setup(gpucontext *c) {
     return error_sys(ctx->err, "calloc");
 
   cuda_enter(ctx);
+  {
+    CUresult err;
+    err = cuCtxGetDevice(&dev);
+    if (err != CUDA_SUCCESS) {
+      cuda_exit(ctx);
+      return error_cuda(ctx->err, "cuCtxGetDevice", err);
+    }
+  }
+  GA_CUDA_EXIT_ON_ERROR(ctx, get_cc(dev, &major, &minor, ctx->err));
+
+  /* Only try to use tensor core on cuda 9 and up */
+  if (ctx->major >= 9 && major >= 7 && minor >= 0) {
+    handle->tensorCore = 1;
+  } else {
+    handle->tensorCore = 0;
+  }
+
   err = cublasCreate(&handle->h);
   if (err != CUBLAS_STATUS_SUCCESS) {
     cuda_exit(ctx);
@@ -254,13 +264,9 @@ static int setup(gpucontext *c) {
   if (e != GA_NO_ERROR) goto e1;
   e = GpuKernel_init(&handle->sgemvBH_T_a1_b1_small, c, 1, &code_sgemvBH_T_a1_b1_small, NULL, "sgemv", 9, types, 0, NULL);
   if (e != GA_NO_ERROR) goto e2;
-  tmp[0] = atomicadd_double;
-  tmp[1] = code_dgemvBH_N_a1_b1_small;
-  e = GpuKernel_init(&handle->dgemvBH_N_a1_b1_small, c, 2, tmp, NULL, "dgemv", 9, types, GA_USE_DOUBLE, NULL);
+  e = GpuKernel_init(&handle->dgemvBH_N_a1_b1_small, c, 1, &code_dgemvBH_N_a1_b1_small, NULL, "dgemv", 9, types, GA_USE_DOUBLE, NULL);
   if (e != GA_NO_ERROR) goto e3;
-  tmp[0] = atomicadd_double;
-  tmp[1] = code_dgemvBH_T_a1_b1_small;
-  e = GpuKernel_init(&handle->dgemvBH_T_a1_b1_small, c, 2, tmp, NULL, "dgemv", 9, types, GA_USE_DOUBLE, NULL);
+  e = GpuKernel_init(&handle->dgemvBH_T_a1_b1_small, c, 1, &code_dgemvBH_T_a1_b1_small, NULL, "dgemv", 9, types, GA_USE_DOUBLE, NULL);
   if (e != GA_NO_ERROR) goto e4;
 
   types[0] = GA_BUFFER;
@@ -276,9 +282,7 @@ static int setup(gpucontext *c) {
   e = GpuKernel_init(&handle->sgerBH_gen_small, c, 1, &code_sgerBH_gen_small, NULL, "_sgerBH_gen_small", 10, types, 0, NULL);
   if (e != GA_NO_ERROR) goto e5;
   types[4] = GA_DOUBLE;
-  tmp[0] = atomicadd_double;
-  tmp[1] = code_dgerBH_gen_small;
-  e = GpuKernel_init(&handle->dgerBH_gen_small, c, 2, tmp, NULL, "_dgerBH_gen_small", 10, types, GA_USE_DOUBLE, NULL);
+  e = GpuKernel_init(&handle->dgerBH_gen_small, c, 1, &code_dgerBH_gen_small, NULL, "_dgerBH_gen_small", 10, types, GA_USE_DOUBLE, NULL);
   if (e != GA_NO_ERROR) goto e6;
 
   ctx->blas_handle = handle;
@@ -460,8 +464,8 @@ static int hgemm(cb_order order, cb_transpose transA, cb_transpose transB,
   ASSERT_BUF(B);
   ASSERT_BUF(C);
 
-  if (cublasSgemmEx == NULL)
-    return error_set(ctx->err, GA_DEVSUP_ERROR, "cublasSgemmEx unavailable");
+  if (cublasSgemmEx == NULL && (cublasGemmEx == NULL || h->tensorCore == 0))
+    return error_set(ctx->err, GA_DEVSUP_ERROR, "cublasSgemmEx|cublasGemmEx unavailable");
 
   if (LARGE_VAL(M) || LARGE_VAL(N) || LARGE_VAL(K) ||
       LARGE_VAL(lda) || LARGE_VAL(ldb) || LARGE_VAL(ldc) ||
@@ -493,15 +497,29 @@ static int hgemm(cb_order order, cb_transpose transA, cb_transpose transB,
   GA_CUDA_EXIT_ON_ERROR(ctx, cuda_wait(B, CUDA_WAIT_READ));
   GA_CUDA_EXIT_ON_ERROR(ctx, cuda_wait(C, CUDA_WAIT_ALL));
 
-  CUBLAS_EXIT_ON_ERROR(ctx, cublasSgemmEx(h->h, convT(transA), convT(transB),
-                                          M, N, K,
-                                          &alpha, ((uint16_t *)A->ptr) + offA,
-                                          CUDA_R_16F,
-                                          lda, ((uint16_t *)B->ptr) + offB,
-                                          CUDA_R_16F,
-                                          ldb, &beta, ((uint16_t *)C->ptr) + offC,
-                                          CUDA_R_16F,
-                                          ldc));
+  if (cublasGemmEx != NULL && h->tensorCore) {
+    CUBLAS_EXIT_ON_ERROR(ctx, cublasGemmEx(h->h, convT(transA), convT(transB),
+					   M, N, K,
+					   &alpha, ((uint16_t *)A->ptr) + offA,
+					   CUDA_R_16F,
+					   lda, ((uint16_t *)B->ptr) + offB,
+					   CUDA_R_16F,
+					   ldb, &beta, ((uint16_t *)C->ptr) + offC,
+					   CUDA_R_16F,
+					   ldc,
+					   CUDA_R_32F,
+					   CUBLAS_GEMM_DFALT_TENSOR_OP));
+  } else {
+    CUBLAS_EXIT_ON_ERROR(ctx, cublasSgemmEx(h->h, convT(transA), convT(transB),
+					    M, N, K,
+					    &alpha, ((uint16_t *)A->ptr) + offA,
+					    CUDA_R_16F,
+					    lda, ((uint16_t *)B->ptr) + offB,
+					    CUDA_R_16F,
+					    ldb, &beta, ((uint16_t *)C->ptr) + offC,
+					    CUDA_R_16F,
+					    ldc));
+  }
 
   GA_CUDA_EXIT_ON_ERROR(ctx, cuda_record(A, CUDA_WAIT_READ));
   GA_CUDA_EXIT_ON_ERROR(ctx, cuda_record(B, CUDA_WAIT_READ));
@@ -852,12 +870,23 @@ static int sgemmBatch(cb_order order, cb_transpose transA, cb_transpose transB,
       return ctx->err->code;
     }
 
+    if (cuda_wait(Ta, CUDA_WAIT_READ) != GA_NO_ERROR) {
+      gpudata_release(Ta);
+      cuda_exit(ctx);
+      return ctx->err->code;
+    }
+
     err = cublasSgemmBatched(h->h,
                              convT(transA), convT(transB),
                              M, N, K, &alpha,
                              (const float **)Aa, lda,
                              (const float **)Ba, ldb, &beta,
                              (float **)Ca, ldc, batchCount);
+    if (cuda_record(Ta, CUDA_WAIT_READ) != GA_NO_ERROR) {
+      gpudata_release(Ta);
+      cuda_exit(ctx);
+      return ctx->err->code;
+    }
     gpudata_release(Ta);
     if (err != CUBLAS_STATUS_SUCCESS) {
       cuda_exit(ctx);
@@ -979,13 +1008,26 @@ static int dgemmBatch(cb_order order, cb_transpose transA, cb_transpose transB,
       return ctx->err->code;
     }
 
+    if (cuda_wait(Ta, CUDA_WAIT_READ) != GA_NO_ERROR) {
+      gpudata_release(Ta);
+      cuda_exit(ctx);
+      return ctx->err->code;
+    }
+
     err = cublasDgemmBatched(h->h,
                              convT(transA), convT(transB),
                              M, N, K, &alpha,
                              (const double **)Aa, lda,
                              (const double **)Ba, ldb, &beta,
                              (double **)Ca, ldc, batchCount);
+
+    if (cuda_record(Ta, CUDA_WAIT_READ) != GA_NO_ERROR) {
+      gpudata_release(Ta);
+      cuda_exit(ctx);
+      return ctx->err->code;
+    }
     gpudata_release(Ta);
+
     if (err != CUBLAS_STATUS_SUCCESS) {
       cuda_exit(ctx);
       return error_cublas(ctx->err, "cublasDgemmBatched", err);
@@ -1653,9 +1695,9 @@ static int sgerBatch(cb_order order, size_t M, size_t N, float alpha,
 
 
   for (i = 0; i < batchCount; i++) {
-    GA_CUDA_EXIT_ON_ERROR(ctx, cuda_record(A[i], CUDA_WAIT_READ));
+    GA_CUDA_EXIT_ON_ERROR(ctx, cuda_record(A[i], CUDA_WAIT_ALL));
     GA_CUDA_EXIT_ON_ERROR(ctx, cuda_record(x[i], CUDA_WAIT_READ));
-    GA_CUDA_EXIT_ON_ERROR(ctx, cuda_record(y[i], CUDA_WAIT_ALL));
+    GA_CUDA_EXIT_ON_ERROR(ctx, cuda_record(y[i], CUDA_WAIT_READ));
   }
 
   cuda_exit(ctx);
@@ -1783,9 +1825,9 @@ static int dgerBatch(cb_order order, size_t M, size_t N, double alpha,
   }
 
   for (i = 0; i < batchCount; i++) {
-    GA_CUDA_EXIT_ON_ERROR(ctx, cuda_record(A[i], CUDA_WAIT_READ));
+    GA_CUDA_EXIT_ON_ERROR(ctx, cuda_record(A[i], CUDA_WAIT_ALL));
     GA_CUDA_EXIT_ON_ERROR(ctx, cuda_record(x[i], CUDA_WAIT_READ));
-    GA_CUDA_EXIT_ON_ERROR(ctx, cuda_record(y[i], CUDA_WAIT_ALL));
+    GA_CUDA_EXIT_ON_ERROR(ctx, cuda_record(y[i], CUDA_WAIT_READ));
   }
 
   cuda_exit(ctx);
