@@ -133,7 +133,7 @@ static int gen_elemwise_basic_kernel(GpuKernel *k, gpucontext *ctx,
   int *ktypes;
   char *size = "ga_size", *ssize = "ga_ssize";
   unsigned int p;
-  int flags = GA_USE_CLUDA;
+  int flags = 0;
   int res;
 
   if (ISSET(gen_flags, GEN_ADDR32)) {
@@ -154,6 +154,7 @@ static int gen_elemwise_basic_kernel(GpuKernel *k, gpucontext *ctx,
 
   p = 0;
 
+  strb_appends(&sb, "#include \"cluda.h\"\n");
   if (preamble)
     strb_appends(&sb, preamble);
   strb_appends(&sb, "\nKERNEL void elem(const ga_size n, ");
@@ -176,7 +177,7 @@ static int gen_elemwise_basic_kernel(GpuKernel *k, gpucontext *ctx,
         ktypes[p++] = GA_SSIZE;
       }
     } else {
-      strb_appendf(&sb, "%s %s", ctype(args[i].typecode), args[j].name);
+      strb_appendf(&sb, "%s %s", ctype(args[j].typecode), args[j].name);
       ktypes[p++] = args[j].typecode;
     }
     if (j != (n - 1)) strb_appends(&sb, ", ");
@@ -212,7 +213,7 @@ static int gen_elemwise_basic_kernel(GpuKernel *k, gpucontext *ctx,
                                         GA_FLOAT : args[j].typecode), args[j].name);
       if (ISSET(args[j].flags, GE_READ)) {
         if (args[j].typecode == GA_HALF && ISSET(gen_flags, GEN_CONVERT_F16)) {
-          strb_appendf(&sb, "%s = load_half((GLOBAL_MEM ga_half *)(((GLOBAL_MEM char *)%s_data) + %s_p));\n",
+          strb_appendf(&sb, "%s = ga_half2float(*(GLOBAL_MEM ga_half *)(((GLOBAL_MEM char *)%s_data) + %s_p));\n",
                        args[j].name, args[j].name, args[j].name);
         } else {
           strb_appendf(&sb, "%s = *(GLOBAL_MEM %s *)(((GLOBAL_MEM char *)%s_data) + %s_p);\n",
@@ -226,7 +227,7 @@ static int gen_elemwise_basic_kernel(GpuKernel *k, gpucontext *ctx,
   for (j = 0; j < n; j++) {
     if (is_array(args[j]) && ISSET(args[j].flags, GE_WRITE)) {
       if (args[j].typecode == GA_HALF && ISSET(gen_flags, GEN_CONVERT_F16)) {
-        strb_appendf(&sb, "store_half((GLOBAL_MEM ga_half *)(((GLOBAL_MEM char *)%s_data) + %s_p), %s);\n",
+        strb_appendf(&sb, "*(GLOBAL_MEM ga_half *)(((GLOBAL_MEM char *)%s_data) + %s_p) = ga_float2half(%s);\n",
                      args[j].name, args[j].name, args[j].name);
       } else {
         strb_appendf(&sb, "*(GLOBAL_MEM %s *)(((GLOBAL_MEM char *)%s_data) + %s_p) = %s;\n",
@@ -276,14 +277,21 @@ static int check_basic(GpuElemwise *ge, void **args, int flags,
   GpuArray *a = NULL, *v;
   unsigned int i, j, p, num_arrays = 0, nd = 0, nnd;
   int call32 = 1;
+  unsigned int nd_i = 0;
+  size_t v_dim_j = 0;
 
   /* Go through the list and grab some info */
   for (i = 0; i < ge->n; i++) {
     if (is_array(ge->args[i])) {
+      nd_i = ((GpuArray *)args[i])->nd;
       if (num_arrays == 0)
-        nd = ((GpuArray *)args[i])->nd;
-      else if (((GpuArray *)args[i])->nd != nd)
-        return error_fmt(ctx->err, GA_VALUE_ERROR, "Arg %u has differing nd = %u", i, ((GpuArray *)args[i])->nd);
+        nd = nd_i;
+      else if (nd_i != nd) {
+        if (flags & GE_PADSHAPE)
+          nd = nd_i > nd ? nd_i : nd;
+        else
+          return error_fmt(ctx->err, GA_VALUE_ERROR, "Arg %u has differing nd = %u", i, nd_i);
+      }
       ++num_arrays;
       if (a == NULL && is_output(ge->args[i]))
         a = (GpuArray *)args[i];
@@ -301,7 +309,7 @@ static int check_basic(GpuElemwise *ge, void **args, int flags,
       return error_sys(ctx->err, "ge_grow");
   }
 
-  /* Now we know that all array arguments have the same number of
+  /* Now we know that all array arguments have at most nd
      dimensions and that the expected output size is the size of a */
 
   /* And copy their initial values in */
@@ -309,7 +317,11 @@ static int check_basic(GpuElemwise *ge, void **args, int flags,
   p = 0;
   for (i = 0; i < ge->n; i++) {
     if (is_array(ge->args[i])) {
-      memcpy(ge->strides[p], ((GpuArray *)args[i])->strides, nd*sizeof(ssize_t));
+      /* Left-pad strides with zero on implicitly broadcasted dimensions */
+      memset(ge->strides[p], 0, nd*sizeof(ssize_t));
+      nd_i = ((GpuArray *)args[i])->nd;
+      memcpy((char *)(ge->strides[p]) + (nd - nd_i)*sizeof(ssize_t),
+             ((GpuArray *)args[i])->strides, nd_i*sizeof(ssize_t));
       p++;
     }
   }
@@ -326,16 +338,23 @@ static int check_basic(GpuElemwise *ge, void **args, int flags,
     for (i = 0; i < ge->n; i++) {
       if (is_array(ge->args[i])) {
         v = (GpuArray *)args[i];
-        if (ge->dims[j] != v->dimensions[j]) {
+        nd_i = v->nd;
+        /* Pad shape with 1 if needed for implicitly broadcasted dimensions
+           and shift if needed */
+        if (j < nd - nd_i)
+          v_dim_j = 1;
+        else
+          v_dim_j = v->dimensions[j - (nd - nd_i)];
+        if (ge->dims[j] != v_dim_j) {
           /* We can't broadcast outputs */
           if (ISCLR(flags, GE_BROADCAST) || is_output(ge->args[i]) ||
-              v->dimensions[j] != 1) {
-            return error_fmt(ctx->err, GA_VALUE_ERROR, "Mismatched dimension %u for input %u (expected %" SPREFIX "u got %" SPREFIX "u)", j, i, ge->dims[j], v->dimensions[j]);
+              v_dim_j != 1) {
+            return error_fmt(ctx->err, GA_VALUE_ERROR, "Mismatched dimension %u for input %u (expected %" SPREFIX "u got %" SPREFIX "u)", j, i, ge->dims[j], v_dim_j);
           }
         }
         /* If the dimension is 1 set the strides to 0 regardless since
            it won't change anything in the non-broadcast case. */
-        if (v->dimensions[j] == 1) {
+        if (v_dim_j == 1) {
           ge->strides[p][j] = 0;
         }
         call32 &= v->offset < ADDR32_MAX;
@@ -433,7 +452,7 @@ static int gen_elemwise_contig_kernel(GpuKernel *k,
   int *ktypes = NULL;
   unsigned int p;
   unsigned int j;
-  int flags = GA_USE_CLUDA;
+  int flags = 0;
   int res;
 
   flags |= gpuarray_type_flagsa(n, args);
@@ -450,6 +469,7 @@ static int gen_elemwise_contig_kernel(GpuKernel *k,
 
   p = 0;
 
+  strb_appends(&sb, "#include \"cluda.h\"\n");
   if (preamble)
     strb_appends(&sb, preamble);
   strb_appends(&sb, "\nKERNEL void elem(const ga_size n, ");
@@ -488,7 +508,7 @@ static int gen_elemwise_contig_kernel(GpuKernel *k,
                                           GA_FLOAT : args[j].typecode), args[j].name);
       if (ISSET(args[j].flags, GE_READ)) {
         if (args[j].typecode == GA_HALF && ISSET(gen_flags, GEN_CONVERT_F16)) {
-          strb_appendf(&sb, "%s = load_half(&%s_p[i]);\n", args[j].name, args[j].name);
+          strb_appendf(&sb, "%s = ga_half2float(%s_p[i]);\n", args[j].name, args[j].name);
         } else {
           strb_appendf(&sb, "%s = %s_p[i];\n", args[j].name, args[j].name);
         }
@@ -502,7 +522,7 @@ static int gen_elemwise_contig_kernel(GpuKernel *k,
     if (is_array(args[j])) {
       if (ISSET(args[j].flags, GE_WRITE)) {
         if (args[j].typecode == GA_HALF && ISSET(gen_flags, GEN_CONVERT_F16)) {
-          strb_appendf(&sb, "store_half(&%s_p[i], %s);\n", args[j].name, args[j].name);
+          strb_appendf(&sb, "%s_p[i] = ga_float2half(%s);\n", args[j].name, args[j].name);
         } else {
           strb_appendf(&sb, "%s_p[i] = %s;\n", args[j].name, args[j].name);
         }
@@ -527,7 +547,6 @@ static int gen_elemwise_contig_kernel(GpuKernel *k,
 static int check_contig(GpuElemwise *ge, void **args,
                         size_t *_n, int *contig) {
   GpuArray *a = NULL, *v;
-  gpucontext *ctx = GpuKernel_context(&ge->k_contig);
   size_t n = 1;
   unsigned int i, j;
   int c_contig = 1, f_contig = 1;
@@ -543,10 +562,10 @@ static int check_contig(GpuElemwise *ge, void **args,
       f_contig &= GpuArray_IS_F_CONTIGUOUS(v);
       if (a != v) {
         if (a->nd != v->nd)
-          return error_fmt(ctx->err, GA_INVALID_ERROR, "Mismatched nd for input %u (expected %u, got %u)", i, a->nd, v->nd);
+          return -1; /* We don't check the value of the error code */
         for (j = 0; j < a->nd; j++) {
           if (v->dimensions[j] != a->dimensions[j])
-            return error_fmt(ctx->err, GA_VALUE_ERROR, "Mismatched dimension %u (expected %" SPREFIX "u, got %" SPREFIX "u)", j, a->dimensions[j], v->dimensions[j]);
+            return -1; /* We don't check the value of the error code */
         }
       }
     }
@@ -719,18 +738,24 @@ GpuElemwise *GpuElemwise_new(gpucontext *ctx,
 
 void GpuElemwise_free(GpuElemwise *ge) {
   unsigned int i;
-  for (i = 0; i < ge->nd; i++) {
-    if (k_initialized(&ge->k_basic_32[i]))
-      GpuKernel_clear(&ge->k_basic_32[i]);
-    if (k_initialized(&ge->k_basic[i]))
-      GpuKernel_clear(&ge->k_basic[i]);
-  }
+  if (ge->k_basic_32 != NULL)
+    for (i = 0; i < ge->nd; i++) {
+      if (k_initialized(&ge->k_basic_32[i]))
+        GpuKernel_clear(&ge->k_basic_32[i]);
+    }
+  if (ge->k_basic != NULL)
+    for (i = 0; i < ge->nd; i++) {
+      if (k_initialized(&ge->k_basic[i]))
+        GpuKernel_clear(&ge->k_basic[i]);
+    }
   if (ge->strides != NULL)
     for (i = 0; i < ge->narray; i++) {
       free(ge->strides[i]);
     }
   if (k_initialized(&ge->k_contig))
     GpuKernel_clear(&ge->k_contig);
+  free(ge->k_basic_32);
+  free(ge->k_basic);
   free_args(ge->n, ge->args);
   free((void *)ge->preamble);
   free((void *)ge->expr);
