@@ -19,7 +19,7 @@ void *ga_load_library(const char *name, error *e) {
 void *ga_func_ptr(void *h, const char *name, error *e) {
   void *res = dlsym(h, name);
   if (res == NULL)
-    error_fmt(e, GA_LOAD_ERROR, "Could not find synbol \"%s\": %s", name, dlerror());
+    error_fmt(e, GA_LOAD_ERROR, "Could not find symbol \"%s\": %s", name, dlerror());
   return res;
 }
 
